"""
This module contains the models for the JAXFIN package.

The submodules of this module are:
    
        * gbm: Geometric Brownian Motion model
"""
from jaxfin.models import gbm

__all__ = ["gbm"]
